from rubikai.solver import util


class SearchProblem:
    """
    This class outlines the structure of a search problem, but doesn't implement
    any of the methods (in object-oriented terminology: an abstract class).

    You do not need to change anything in this class, ever.
    """

    def get_start_state(self):
        """
        Returns the start state for the search problem
        """
        util.raiseNotDefined()

    def is_goal_state(self, state):
        """
        state: Search state

        Returns True if and only if the state is a valid goal state
        """
        util.raiseNotDefined()

    def get_successors(self, state):
        """
        state: Search state

        For a given state, this should return a list of triples,
        (successor, action, stepCost), where 'successor' is a
        successor to the current state, 'action' is the action
        required to get there, and 'stepCost' is the incremental
        cost of expanding to that successor
        """
        util.raiseNotDefined()

    def get_cost_of_actions(self, actions):
        """
        actions: A list of actions to take

        This method returns the total cost of a particular sequence of actions.  The sequence must
        be composed of legal moves
        """
        util.raiseNotDefined()


class Node:
    """
    A simple node class which saves arbitrary data and a pointer to a parent.
    """

    def __init__(self, data, parent=None):
        """
        Initializes a new node
        :param data: data to save
        :param parent: node's parent (Node object)
        """
        self.data = data
        self.parent = parent

    def is_root(self):
        """
        :return: True if this node has a parent, False otherwise
        """
        return True if self.parent is None else False

    def path_from_root(self):
        """
        :return: a list of nodes which corresponds to a path from the root
                 to this node
        """
        path = []
        current = self
        while current is not None:
            path.append(current)
            current = current.parent
        # the above path is reversed. return the correct order:
        return path[::-1]


class SearchNode(Node):
    """ a node where the data saved is the (state,action,cost) triplet """
    def __init__(self, state, action, cost, parent=None):
        super().__init__(None, parent)
        self.state = state
        self.action = action
        self.cost = cost


def _cost_so_far(node):
    """ also referred to as `g' in the class notation """
    prev_cost = 0 if node.is_root() else node.parent.incremental_cost
    node.incremental_cost = prev_cost + node.cost
    return node.incremental_cost


def generic_search(problem, data_structure, **kwargs):
    """
    Implements a generic search algorithm which. I.e. this algorithm
    can be used with different data structures as fringe.

    :param problem: The search problem instance
    :param data_structure: A data structure for the fringe (e.g. Stack)
    :param kwargs: Additional keyword arguments for the data structure init
    :return: A list of actions if a goal state is found, None otherwise
    """

    fringe = data_structure(**kwargs)
    visited = set()
    start_state = problem.get_start_state()
    fringe.push(SearchNode(state=start_state, action=None, cost=0))
    visited.add(start_state)
    while not fringe.is_empty():
        node = fringe.pop()
        if problem.is_goal_state(node.state):
            # omit the root node (doesn't have any action)
            path_from_root = node.path_from_root()[1:]
            return [p.action for p in path_from_root]
        else:
            visited.add(node.state)
            for successor, action, cost in problem.get_successors(node.state):
                if successor not in visited:
                    fringe.push(SearchNode(successor, action,
                                           cost, parent=node))
    # This means that the fringe is empty and no goal state was reached
    return []


def depth_first_search(problem):
    """
    Search the deepest nodes in the search tree first.
    """
    return generic_search(problem, util.Stack)


def breadth_first_search(problem):
    """
    Search the shallowest nodes in the search tree first.
    """
    return generic_search(problem, util.Queue)


def uniform_cost_search(problem):
    """
    Search the node of least total cost first.
    """
    return generic_search(problem, util.PriorityQueueWithFunction,
                          priority_function=_cost_so_far)


def null_heuristic(*_):
    """
    A heuristic function estimates the cost from the current state to the nearest
    goal in the provided SearchProblem.  This heuristic is trivial.
    """
    return 0


def a_star_search(problem, heuristic=null_heuristic):
    """
    Search the node that has the lowest combined cost and heuristic first.
    """
    def f(n):
        return heuristic(n.state, problem) + _cost_so_far(n)

    return generic_search(problem, util.PriorityQueueWithFunction,
                          priority_function=f)


# Abbreviations
bfs = breadth_first_search
dfs = depth_first_search
a_star = a_star_search
ucs = uniform_cost_search



from rubikai.solver import search
from rubikai.cube.cube import Cube, Face, Action
import numpy as np
import pandas as pd


class CubeProblem(search.SearchProblem):
    """ a Problem class to be used with the search module """

    def __init__(self, cube, quarter_metric=True):
        """
        :param cube: a rubikai.cube.cube.Cube instance
        :param quarter_metric: if True, use 90 degrees rotations only.
                               otherwise use also 180 degrees.
        """
        self.cube = cube
        self.expanded = 0
        self.actions = []
        if quarter_metric:
            k_values = (-1, 1)
        else:
            k_values = (-1, 1, 2)
        for layer in range(self.cube.layers // 2):
            for k in k_values:
                self.actions += [Action(face, k, layer) for face in Face]

    def get_start_state(self):
        return self.cube.copy()

    def is_goal_state(self, cube):
        return cube.is_solved()

    def get_successors(self, cube):
        self.expanded += 1
        successors = []
        for action in self.actions:
            successor = cube.copy().rotate(
                action.face, action.k, action.layer)
            successors.append((successor, action, 1))
        return successors

    def get_cost_of_actions(self, actions):
        return len(actions)


def solve(cube, heuristic=lambda *args: 0, verbose=False):
    """
    solves a given cube instance
    :param cube: rubikai.cube.cube.Cube instance
    :param heuristic: a heuristic function for A* search.
                      takes a cube and problem instances as input and returns
                      a number.
                      by default, uses the null heuristic (0 for every state)
    :param verbose: if True, prints some information about the search
    :return: a list of actions that solves the cube
    """
    cube_problem = CubeProblem(cube)
    solution = search.a_star(cube_problem, heuristic)
    if verbose:
        solution_str = ' '.join([str(s) for s in solution])
        print('Length %d solution found:\n%s' %
              (len(solution), solution_str))
        print('Expanded %d nodes.' % cube_problem.expanded)
    return solution, cube_problem.expanded


def random_actions(cube, num_actions, quarter_metric=True):
    """
    returns a sequence of random actions on the given cube
    (does not apply the actions to the cube)
    :param cube: Cube object
    :param num_actions: an integer >= 1
    :param quarter_metric: whether to use the quarter-turn metric or not
    :return: a sequence of `num_actions' actions
    """
    assert num_actions >= 1, "there's no sense in less than 1 steps"
    cube_problem = CubeProblem(cube, quarter_metric)
    last_cube = cube.copy()
    visited = {last_cube.copy()}
    possible_actions = cube_problem.actions
    actions = [None] * num_actions
    actions[0] = np.random.choice(possible_actions)
    for i in range(1, num_actions):
        rand_action = np.random.choice(possible_actions)
        while last_cube.copy().apply([rand_action]) in visited:
            rand_action = np.random.choice(possible_actions)
        actions[i] = rand_action
        last_cube.apply([rand_action])
        visited.add(last_cube.copy())
    return actions


def switchable(a_1, a_2):
    """
    decide if 2 actions can be swapped.
    two actions are independent if they are on opposite faces or on the same
    face but on different layers.
    :param a_1: first action
    :param a_2: second action
    :returns True if switchable, False otherwise
    """
    opposite_faces = a_1.face.value + a_2.face.value == 5
    same_face = a_1.face.value == a_2.face.value
    diff_layers = a_1.layer != a_2.layer
    if opposite_faces:
        return a_1.face.value > a_2.face.value
    elif same_face and diff_layers:
        return a_1.layer > a_2.layer
    else:
        return False


def bubble_sort(actions):
    """ bubble sort actions were switching is between independent actions """
    exchanges = True
    pass_num = len(actions) - 1
    while pass_num > 0 and exchanges:
        exchanges = False
        for i in range(pass_num):
            if switchable(actions[i], actions[i + 1]):
                exchanges = True
                temp = actions[i]
                actions[i] = actions[i + 1]
                actions[i + 1] = temp
        pass_num = pass_num - 1
    return actions


def reduce_same_actions(actions):
    """ reduce a sequence of the same actions """
    reduced_actions = []
    i = 0
    while i < len(actions):
        curr_action_k = actions[i].k
        j = i + 1
        while j < len(actions) and \
                actions[i].face.value == actions[j].face.value and \
                actions[i].layer == actions[j].layer:
            curr_action_k += actions[j].k
            j += 1
        # end while
        curr_action_k %= 4
        if curr_action_k == 2:
            a = Action(actions[i].face, -1, actions[i].layer)
            reduced_actions.append(a)
            reduced_actions.append(a)
        elif curr_action_k != 0:
            a = Action(actions[i].face, curr_action_k, actions[i].layer)
            reduced_actions.append(a)
        # end if
        i = j
    # end while
    return reduced_actions


def reduce_sequence(actions):
    """
    two actions on the cube are independent if they don't affect the same faces,
    in this case we can change the order of the actions.
    we then can go over sequences of dependent action and change a series
    of action to shorter ones.
    """
    return reduce_same_actions(bubble_sort(actions))


def generate_random_sequence(cube_size, d):
    if d == 0:
        return []
    rand_actions = []
    happy = False
    while not happy:
        cube = Cube(cube_size)
        rand_actions = random_actions(cube, 2*d)
        rand_actions = reduce_sequence(rand_actions)
        happy = len(rand_actions) >= d
    return rand_actions[:d]


def compare_heuristics(heuristics, cube_layers, d_values,
                       iterations, verbose=False):
    """
    uses the given heuristics to solve multiple instances of scrambled cubes
    and gathers information on the solutions.

    :param heuristics: a dictionary where the key is the heuristic name and the
                       value is the heuristic function handle
    :param cube_layers: number of cube layers the heuristic operates on
    :param d_values: an array-like of integers specifying the number of
                     scramble moves to check
    :param iterations: number of iterations for each num. of scramble moves
    :param verbose: if True, prints information as the test goes
    :returns: pandas DataFrame with all the data gathered
    """

    # print only if verbose is True
    def vprint(*a, **kw):
        if verbose:
            print(*a, **kw)
    # end vprint
    # create a dataframe
    name_col = 'heuristic_name'
    d_col = 'num_scrambles'
    opt_col = 'is_optimal'
    exp_col = 'expansions'
    df = pd.DataFrame(columns=[name_col, d_col, opt_col, exp_col])

    for i in range(iterations):
        vprint('Iteration:', i, '\n')
        for d in d_values:
            # generate a random sequence of length d and scramble the cube
            rand_seq = generate_random_sequence(cube_layers, d)
            vprint('Generated sequence:', rand_seq, '(length %d)' % d)
            c = Cube(cube_layers)
            c.apply(rand_seq)
            # test each heuristic on the scrambled cube
            for h_name in heuristics:
                h = heuristics[h_name]
                solution, expansions = solve(c, h, verbose)
                df = df.append({
                    name_col: h_name,
                    d_col: d,
                    opt_col: len(solution) == d,
                    exp_col: expansions
                }, ignore_index=True)
                vprint()
            # end for
            vprint()
        # end for
        vprint()
    # end for
    return df



from rubikai.solver.search import dfs, bfs, ucs, a_star, null_heuristic
from rubikai.cube.cube import Cube, Face, Action, NUM_FACES
from rubikai.solver.solver import solve, random_actions, \
    generate_random_sequence, compare_heuristics
from rubikai.solver.learning import scrambled_cube_generator, batch_generator

name = "rubikai"



# util.py
# -------
# Licensing Information: Please do not distribute or publish solutions to this
# project. You are free to use and extend these projects for educational
# purposes. The Pacman AI projects were developed at UC Berkeley, primarily by
# John DeNero (denero@cs.berkeley.edu) and Dan Klein (klein@cs.berkeley.edu).
# For more info, see http://inst.eecs.berkeley.edu/~cs188/sp09/pacman.html

import heapq

"""
 Data structures useful for implementing SearchAgents
"""


class Stack:
    """"A container with a last-in-first-out (LIFO) queuing policy."""
    def __init__(self):
        self.list = []

    def push(self, item):
        """"Push 'item' onto the stack"""
        self.list.append(item)

    def pop(self):
        """"Pop the most recently pushed item from the stack"""
        return self.list.pop()

    def is_empty(self):
        """"Returns true if the stack is empty"""
        return len(self.list) == 0


class Queue:
    """"A container with a first-in-first-out (FIFO) queuing policy."""
    def __init__(self):
        self.list = []

    def push(self, item):
        """"Enqueue the 'item' into the queue"""
        self.list.insert(0, item)

    def pop(self):
        """
          Dequeue the earliest enqueued item still in the queue. This
          operation removes the item from the queue.
        """
        return self.list.pop()

    def is_empty(self):
        """"Returns true if the queue is empty"""
        return len(self.list) == 0


class PriorityQueue:
    """
      Implements a priority queue data structure. Each inserted item
      has a priority associated with it and the client is usually interested
      in quick retrieval of the lowest-priority item in the queue. This
      data structure allows O(1) access to the lowest-priority item.

      Note that this PriorityQueue does not allow you to change the priority
      of an item.  However, you may insert the same item multiple times with
      different priorities.
    """
    def __init__(self):
        self.heap = []
        self.init = False

    def push(self, item, priority):
        if not self.init:
            self.init = True
            try:
                item < item
            except:
                item.__class__.__lt__ = lambda x, y: True
        pair = (priority, item)
        heapq.heappush(self.heap, pair)

    def pop(self):
        (priority, item) = heapq.heappop(self.heap)
        return item

    def is_empty(self):
        return len(self.heap) == 0


class PriorityQueueWithFunction(PriorityQueue):
    """
    Implements a priority queue with the same push/pop signature of the
    Queue and the Stack classes. This is designed for drop-in replacement for
    those two classes. The caller has to provide a priority function, which
    extracts each item's priority.
    """
    def __init__(self, priority_function):
        """"priority_function (item) -> priority"""
        self.priority_function = priority_function  # store the priority function
        super().__init__()  # super-class initializer

    def push(self, item):
        """"Adds an item to the queue with priority from the priority function"""
        PriorityQueue.push(self, item, self.priority_function(item))


def manhattan_distance(xy1, xy2):
    """"Returns the Manhattan distance between points xy1 and xy2 in R^2"""
    return abs(xy1[0] - xy2[0]) + abs(xy1[1] - xy2[1])



import numpy as np
from rubikai.cube.cube import Cube
from rubikai.solver.solver import generate_random_sequence


def scrambled_cube_generator(layers, max_d, p=None):
    """
    a generator that performs a random walk _back_ from a solved cube, where
    the number of moves performed is at most `max_d' (inclusive).
    yields a pair of a scrambled_cube_array and the number of actual moves
    performed.
    :param layers: the number of layers in the scrambled cube
    :param max_d: the maximal depth of the random walk
    :param p: (optional) a length (max_d+1) probabilities vector according to
              which the number of scramble-steps is chosen.
    :return: (arr, d) pairs, where `arr' is a scrambled-cube array and d is
              the number of moves that were performed to scramble it.
    """
    d_values = np.arange(max_d + 1)
    while True:
        cube = Cube(layers)
        d = np.random.choice(d_values, p=p)
        actions = generate_random_sequence(layers, d)
        cube.apply(actions)
        yield cube.to_array(), d


def batch_generator(layers, max_d, batch_size, p=None):
    """
    same as `scrambled_cube_generator', but for batches of data.
    :param layers: cube dimension (layers x layers x layers cube)
    :param max_d: maximal number of scramble-steps
    :param batch_size: number of instances for each iteration
    :param p: num-of-steps probabilities vector (see scrambled_cube_generator)
    :return: (data, labels) pair, where `data' is a 2d array in which each row
              is a scrambled cube array. labels is a 1d array where each entry
              corresponds to a row in the data array.
    """
    dimension = len(Cube(layers).to_array())
    while True:
        # preallocate the output arrays
        data = np.empty((batch_size, dimension), dtype=np.int8)
        labels = np.empty(batch_size, dtype=np.int8)
        # generate instances until batch_size is reached
        for i, (arr, d) in enumerate(
                scrambled_cube_generator(layers, max_d, p)):
            if i >= batch_size:
                break
            data[i, :] = arr
            labels[i] = d
        yield data, labels



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rubikai",
    version="0.1.2",
    author="Forest Agostinelli",
    author_email="agostinelli.f@hotmail.com",
    description="Rubik AI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy',
        'pandas'
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",  # choose a license
        "Operating System :: OS Independent",
    ),
)



# imports
import numpy as np
from enum import Enum


# helper functions
def invert_permutation(p):
    """ invert a permutation given by the list `p' """
    inv = np.empty(len(p), dtype=int)
    for i, val in enumerate(p):
        inv[val] = i
    return inv


class Cubbie(Enum):
    """
      O
    G W B Y
      R
    """
    W = 0
    R = 1
    B = 2
    G = 3
    O = 4
    Y = 5


CUBBIE_TO_STR = {
    Cubbie.W:  'W',
    Cubbie.G:  'G',
    Cubbie.R:  'R',
    Cubbie.Y:  'Y',
    Cubbie.B:  'B',
    Cubbie.O:  'O'
}

STR_TO_CUBBIE = {CUBBIE_TO_STR[k]: k for k in CUBBIE_TO_STR.keys()}


def num_to_color(num):
    # get the relevant cubbie and transform it to a string
    cubbie = [c for c in Cubbie if c.value == num][0]
    cubbie_str = CUBBIE_TO_STR[cubbie]
    # add background color
    color_str = '\033['
    if num == 0:  # W
        color_str += '47m'
    elif num == 1:  # R
        color_str += '41m'
    elif num == 2:  # B
        color_str += '44m'
    elif num == 3:  # G
        color_str += '42m'
    elif num == 4:  # O
        color_str += '45m'
    elif num == 5:  # Y
        color_str += '43m'
    # change text color
    if num == 0:
        color_str += '\033[30m'
    else:
        color_str += '\033[37m'
    return color_str + cubbie_str + '\033[0m'

# constants
NUM_FACES = 6


class Face(Enum):
    """ Enumeration of the cube's faces """
    FRONT = 0
    DOWN = 1
    RIGHT = 2
    LEFT = 3
    UP = 4
    BACK = 5


# go from a string to a face and backwards
FACE_TO_STR = {
    Face.FRONT: 'F',
    Face.BACK: 'B',
    Face.RIGHT: 'R',
    Face.LEFT: 'L',
    Face.UP: 'U',
    Face.DOWN: 'D'
}
STR_TO_FACE = {FACE_TO_STR[k]: k for k in FACE_TO_STR.keys()}

# the permutations and rotations for the Cube.rotate method
PERMUTATION = {
    Face.FRONT: [0, 1, 2, 3, 4, 5],
    Face.DOWN: [1, 5, 2, 3, 0, 4],
    Face.RIGHT: [2, 1, 5, 0, 4, 3],
    Face.LEFT: [3, 1, 0, 5, 4, 2],
    Face.UP: [4, 0, 2, 3, 5, 1],
    Face.BACK: [5, 1, 3, 2, 4, 0]
}
INVERSE_PERM = {face: invert_permutation(PERMUTATION[face])
                for face in Face}
ROTATION = {
    Face.FRONT: [0, 0, 0, 0, 0, 0],
    Face.DOWN: [0, 0, -1, 1, 2, 2],
    Face.RIGHT: [0, 1, 0, 0, -1, 0],
    Face.LEFT: [0, -1, 0, 0, 1, 0],
    Face.UP: [0, 2, 1, -1, 0, 2],
    Face.BACK: [0, 2, 0, 0, 2, 0]
}


# class definitions
class Action:
    """ a class to nicely represent an action. """

    def __init__(self, face_or_str, k=1, layer=0):
        """
        can be initialized either with a standard string (e.g. "F'") or
        explicitly give a face, amount and layer.
        :param face_or_str: rubikai.cube.cube.Face or string
        :param k: int
        :param layer: non-negative int
        """
        if isinstance(face_or_str, str):
            self.__init_from_str__(face_or_str)
        else:
            self.face = face_or_str
            self.k = k
            self.layer = layer

    def __init_from_str__(self, string):
        """
        initializes the action from an action string
        :param string: F B L R U D for front, back, left, right, up, down
                       respectively. default is clockwise rotation.
                       add ' in the end (e.g. R') for counter-clockwise,
                       or 2 for 180 degrees rotation (e.g. R2)
        """
        assert 1 <= len(string) <= 2
        assert string[0].upper() in ['F', 'B', 'R', 'L', 'U', 'D']
        self.face = STR_TO_FACE[string[0].upper()]
        if string[0].isupper():
            self.layer = 0
        else:
            self.layer = 1

        if string[-1] == "'":
            self.k = 1
        elif string[-1] == "2":
            self.k = 2
        else:
            self.k = 3

    def inverse(self):
        face = self.face
        k = (-self.k) % 4
        layer = self.layer
        return Action(face, k, layer)

    def __eq__(self, other):
        if isinstance(other, str):
            return self.__eq__(Action(other))
        return self.face == other.face \
            and self.k == other.k \
            and self.layer == other.k

    def __str__(self):
        result = FACE_TO_STR[self.face]
        if self.layer == 1:
            result = result.lower()
        if self.k == 1:
            result += "'"
        elif self.k == 2:
            result += "2"
        return result

    def __repr__(self):
        return str(self)


class Cube:
    """ A simple representation of a NxNxN cube (3x3x3 by default) """

    def __init__(self, layers=3):
        """
        Creates a new instance of a NxNxN cube, where N is the `layers' param
        :param layers: the number of layers in the cube (3 by default)
        """
        assert layers >= 2, "a cube must have at least 2 layers"
        self.layers = layers
        self._array = np.zeros((layers, layers, NUM_FACES),
                               dtype=np.uint8)
        for face in Face:
            self._array[:, :, face.value] = face.value

    def copy(self):
        new_cube = Cube(self.layers)
        new_cube._array = self._array.copy()
        return new_cube

    @staticmethod
    def _rotate_front(arr, amount, layer):
        """
        Rotates the `layer'th layer in the front face of the cube represented
        by the array `arr'.
        :param arr: NxNx6 array representing a cube
        :param amount: the number of 90 deg. rotations clockwise
        :param layer: which layer to rotate (int between 0 and the number
                      of layers in the cube divided by 2)
        :return: a rotated array
        """
        
        def v(x):
            """ shorthand for the long faces names """
            return STR_TO_FACE[x].value
        
        arr_copy = arr.copy()
        if layer == 0:
            arr_copy[:, :, v('F')] = np.rot90(arr[:, :, v('F')], amount)
        for _ in range(amount):
            arr_copy[-(1+layer), :, v('U')] = arr[:, layer, v('R')]
            arr_copy[:, layer, v('R')] = arr[layer, :, v('D')][::-1]
            arr_copy[layer, :, v('D')] = arr[:, -(1+layer), v('L')]
            arr_copy[:, -(1+layer), v('L')] = arr[-(1+layer), :, v('U')][::-1]
            arr = arr_copy.copy()
        return arr_copy

    def to_array(self, flat=True):
        """
        :param flat: True for a flat array (1d), False for NxNx6 array
        :return: the array representing this cube
        """
        if flat:
            return self._array.flatten()
        else:
            return self._array.copy()

    def rotate(self, face, amount, layer=0):
        """
        Rotates a given layer parallel to a face of the cube
        :param face: must be a `Face' instance
        :param amount: number of counterclockwise rotations
        :param layer: which layer to rotate (int in the range [0,m/2) where
                      m is the number of layers in the cube)
        """
        # make sure the input is valid
        assert isinstance(face, Face), "the `face' parameter must be a \
            Face object"
        assert 0 <= layer < (self.layers // 2), "the layer must be smaller \
            than: dimension / 2"

        amount %= 4
        if amount == 0:
            return self
        # get the relevant permutations and rotations
        perm = PERMUTATION[face]
        rotation = ROTATION[face]
        inverse = INVERSE_PERM[face]
        rotated = self._array.copy()
        # change the cube's orientation such that the input face is
        # the front face
        for i in range(NUM_FACES):
            rotated[:, :, i] = np.rot90(rotated[:, :, i], rotation[i])
        rotated = rotated[:, :, perm]
        # rotate the front (which is now the correct face to rotate)
        rotated = Cube._rotate_front(rotated, amount, layer)
        # change the cube back to the original orientation
        rotated = rotated[:, :, inverse]
        for i in range(NUM_FACES):
            rotated[:, :, i] = np.rot90(rotated[:, :, i], -rotation[i])
        # save the result in the inner array
        self._array = rotated
        return self

    def apply(self, sequence):
        # if the sequence is empty, do nothing
        if not sequence:
            return self
        # handle a string (e.g. "U D' R2 B")
        elif isinstance(sequence, str):
            return self.apply(list(Action(a) for a in sequence.split()))
        # handle a list of strings (e.g. ["U", "D'", "R2", "B"])
        elif isinstance(sequence[0], str):
            return self.apply(list(Action(a) for a in sequence))
        # handle a list of actions
        else:
            for action in sequence:
                self.rotate(action.face, action.k, action.layer)
            return self

    def is_solved(self):
        """
        :return: True if the cube is solved, False otherwise
        """
        return all(np.all(self._array[:, :, i] == self._array[0, 0, i])
                   for i in range(NUM_FACES))

    def get_face(self, face):
        """
        Get a 2d array of the chosen face
        :param face: a Face instance
        :return: NxN array of colors of the chosen face
        """
        assert isinstance(face, Face), "the 'face' parameter must be a Face object"
        return self._array[:, :, face.value].copy()

    def __str__(self):
        result = ''
        n = self.layers
        blank = [' '] * n
        block_sep = ' '
        line_sep = '\n'
        cubbie_sep = ''

        def row(arr):
            if arr[0] != ' ':
                arr = [num_to_color(arr[i]) for i in range(n)]
            return ''.join([arr[i] + cubbie_sep for i in range(n)])

        for i in range(n):
            result += row(blank) + block_sep + \
                      row(self._array[i, :, Face.UP.value]) + '\n'
        result += line_sep
        for i in range(n):
            result += row(self._array[i, :, Face.LEFT.value]) + block_sep + \
                row(self._array[i, :, Face.FRONT.value]) + block_sep + \
                row(self._array[i, :, Face.RIGHT.value]) + block_sep + \
                row(self._array[i, :, Face.BACK.value]) + '\n'
        result += line_sep
        for i in range(n):
            result += row(blank) + block_sep + \
                      row(self._array[i, :, Face.DOWN.value]) + '\n'
        return result

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return np.array_equal(self._array, other._array)

    def set_face(self, face, array):
        assert isinstance(face, Face), "the `face' parameter must be a Face object"
        assert array[1][1] == face.value, "center cube doesn't match face: see instructions on how to enter cube\n"
        self._array[:, :, face.value] = array


def _face_convert_str(array):
    new_array = array.copy()
    assert len(array) == len(array[0]), """ input not valid: incorrect sizes of array """
    for i in range(len(array)):
        for j in range(len(array)):
            assert array[i][j] in STR_TO_CUBBIE.keys(), "input color not valid: view instructions\n"
            new_array[i][j] = STR_TO_CUBBIE[array[i][j]].value
    return new_array


def convert_real_cube_3x3(front, right, left, back, up, down):
    """
    front face with middle cube white, right face with middle cube blue ,the rest are accordingly.
    Enter each face as a 3 by 3 array, each cubbie should be a char of the first letter of the cubbie's color.
    """
    cube = Cube()
    cube.set_face(Face.FRONT, _face_convert_str(front))
    cube.set_face(Face.RIGHT, _face_convert_str(right))
    cube.set_face(Face.LEFT, _face_convert_str(left))
    cube.set_face(Face.BACK, _face_convert_str(back))
    cube.set_face(Face.UP, _face_convert_str(up))
    cube.set_face(Face.DOWN, _face_convert_str(down))
    return cube

